# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

from collections import defaultdict
import math
import logging

import numpy as num
import matplotlib
from matplotlib.axes import Axes
# from matplotlib.ticker import MultipleLocator
from matplotlib import cm, colors, colorbar, figure

from pyrockoeost.guts import Tuple, Float, Object
from pyrockoeost import plot

import scipy.optimize

logger = logging.getLogger('pyrockoeost.plot.smartplot')

guts_prefix = 'pf'

inch = 2.54


class SmartplotAxes(Axes):

    if matplotlib.__version__.split('.') < '3.6'.split('.'):
        # Subclassing cla is deprecated on newer mpl but need this fallback for
        # older versions. Code is duplicated because mpl behaviour depends
        # on the existence of cla in the subclass...
        def cla(self):
            if hasattr(self, 'callbacks'):
                callbacks = self.callbacks
                Axes.cla(self)
                self.callbacks = callbacks
            else:
                Axes.cla(self)

    else:
        def clear(self):
            if hasattr(self, 'callbacks'):
                callbacks = self.callbacks
                Axes.clear(self)
                self.callbacks = callbacks
            else:
                Axes.clear(self)


class SmartplotFigure(figure.Figure):

    def set_smartplot(self, plot):
        self._smartplot = plot

    def draw(self, *args, **kwargs):
        if hasattr(self, '_smartplot'):
            try:
                self._smartplot._update_layout()
            except NotEnoughSpace:
                logger.error('Figure is too small to show the plot.')
                return

        return figure.Figure.draw(self, *args, **kwargs)


def limits(points):
    lims = num.zeros((3, 2))
    if points.size != 0:
        lims[:, 0] = num.min(points, axis=0)
        lims[:, 1] = num.max(points, axis=0)

    return lims


def wcenter(rect):
    return rect[0] + rect[2]*0.5


def hcenter(rect):
    return rect[1] + rect[3]*0.5


def window_min(n, w, ml, mu, s, x):
    return ml + x/float(n) * (w - (ml + mu + (n-1)*s)) + math.floor(x) * s


def window_max(n, w, ml, mu, s, x):
    return ml + x/float(n) * (w - (ml + mu + (n-1)*s)) + (math.floor(x)-1) * s


def make_smap(cmap, norm=None):
    if isinstance(norm, tuple):
        norm = colors.Normalize(*norm, clip=False)
    smap = cm.ScalarMappable(cmap=cmap, norm=norm)
    smap._A = []  # not needed in newer versions of mpl?
    return smap


def solve_layout_fixed_panels(size, shape, limits, aspects, fracs=None):

    weight_aspect = 1000.

    sx, sy = size
    nx, ny = shape
    nvar = nx+ny
    vxs, vys = limits
    uxs = vxs[:, 1] - vxs[:, 0]
    uys = vys[:, 1] - vys[:, 0]
    aspects_xx, aspects_yy, aspects_xy = aspects

    if fracs is None:
        wxs = num.full(nx, sx / nx)
        wys = num.full(ny, sy / ny)
    else:
        frac_x, frac_y = fracs
        wxs = sx * frac_x / num.sum(frac_x)
        wys = sy * frac_y / num.sum(frac_y)

    data = []
    weights = []
    rows = []
    bounds = []
    for ix in range(nx):
        u = uxs[ix]
        assert u > 0.0
        row = num.zeros(nvar)
        row[ix] = u
        rows.append(row)
        data.append(wxs[ix])
        weights.append(1.0 / u)
        bounds.append((0, wxs[ix] / u))

    for iy in range(ny):
        u = uys[iy]
        assert u > 0.0
        row = num.zeros(nvar)
        row[nx+iy] = u
        rows.append(row)
        data.append(wys[iy])
        weights.append(1.0)
        bounds.append((0, wys[iy] / u))

    for ix1, ix2, aspect in aspects_xx:
        row = num.zeros(nvar)
        row[ix1] = aspect
        row[ix2] = -1.0
        weights.append(weight_aspect/aspect)
        rows.append(row)
        data.append(0.0)

    for iy1, iy2, aspect in aspects_yy:
        row = num.zeros(nvar)
        row[nx+iy1] = aspect
        row[nx+iy2] = -1.0
        weights.append(weight_aspect/aspect)
        rows.append(row)
        data.append(0.0)

    for ix, iy, aspect in aspects_xy:
        row = num.zeros(nvar)
        row[ix] = aspect
        row[nx+iy] = -1.0
        weights.append(weight_aspect/aspect)
        rows.append(row)
        data.append(0.0)

    weights = num.array(weights)
    data = num.array(data)
    mat = num.vstack(rows) * weights[:, num.newaxis]
    data *= weights

    bounds = num.array(bounds).T

    model = scipy.optimize.lsq_linear(mat, data, bounds).x

    cxs = model[:nx]
    cys = model[nx:nx+ny]

    vlimits_x = num.zeros((nx, 2))
    for ix in range(nx):
        u = wxs[ix] / cxs[ix]
        vmin, vmax = vxs[ix]
        udata = vmax - vmin
        eps = 1e-7 * u
        assert udata <= u + eps
        vlimits_x[ix, 0] = (vmin + vmax) / 2.0 - u / 2.0
        vlimits_x[ix, 1] = (vmin + vmax) / 2.0 + u / 2.0

    vlimits_y = num.zeros((ny, 2))
    for iy in range(ny):
        u = wys[iy] / cys[iy]
        vmin, vmax = vys[iy]
        udata = vmax - vmin
        eps = 1e-7 * u
        assert udata <= u + eps
        vlimits_y[iy, 0] = (vmin + vmax) / 2.0 - u / 2.0
        vlimits_y[iy, 1] = (vmin + vmax) / 2.0 + u / 2.0

    def check_aspect(a, awant, eps=1e-2):
        if abs(1.0 - (a/awant)) > eps:
            logger.error(
                'Unable to comply with requested aspect ratio '
                '(wanted: %g, achieved: %g)' % (awant, a))

    for ix1, ix2, aspect in aspects_xx:
        check_aspect(cxs[ix2] / cxs[ix1], aspect)

    for iy1, iy2, aspect in aspects_yy:
        check_aspect(cys[iy2] / cys[iy1], aspect)

    for ix, iy, aspect in aspects_xy:
        check_aspect(cys[iy] / cxs[ix], aspect)

    return (vlimits_x, vlimits_y), (wxs, wys)


def solve_layout_iterative(size, shape, limits, aspects, niterations=3):

    sx, sy = size
    nx, ny = shape
    vxs, vys = limits
    uxs = vxs[:, 1] - vxs[:, 0]
    uys = vys[:, 1] - vys[:, 0]
    aspects_xx, aspects_yy, aspects_xy = aspects

    fracs_x, fracs_y = num.ones(nx), num.ones(ny)
    for i in range(niterations):
        (vlimits_x, vlimits_y), (wxs, wys) = solve_layout_fixed_panels(
            size, shape, limits, aspects, (fracs_x, fracs_y))

        uxs_view = vlimits_x[:, 1] - vlimits_x[:, 0]
        uys_view = vlimits_y[:, 1] - vlimits_y[:, 0]
        wxs_used = wxs * uxs / uxs_view
        wys_used = wys * uys / uys_view
        # wxs_wasted = wxs * (1.0 - uxs / uxs_view)
        # wys_wasted = wys * (1.0 - uys / uys_view)

        fracs_x = wxs_used
        fracs_y = wys_used

    return (vlimits_x, vlimits_y), (wxs, wys)


class PlotError(Exception):
    pass


class NotEnoughSpace(PlotError):
    pass


class PlotConfig(Object):

    font_size = Float.T(default=9.0)

    size_cm = Tuple.T(
        2, Float.T(), default=(20., 20.))

    margins_em = Tuple.T(
        4, Float.T(), default=(8., 6., 8., 6.))

    separator_em = Float.T(default=1.5)

    colorbar_width_em = Float.T(default=2.0)

    label_offset_em = Tuple.T(
        2, Float.T(), default=(2., 2.))

    tick_label_offset_em = Tuple.T(
        2, Float.T(), default=(0.5, 0.5))

    @property
    def size_inch(self):
        return self.size_cm[0]/inch, self.size_cm[1]/inch


class Plot(object):

    def __init__(
            self, x_dims=['x'], y_dims=['y'], z_dims=[], config=None,
            fig=None, call_mpl_init=True):

        if config is None:
            config = PlotConfig()

        self._shape = len(x_dims), len(y_dims)

        dims = []
        for dim in x_dims + y_dims + z_dims:
            dim = dim.lstrip('-')
            if dim not in dims:
                dims.append(dim)

        self.config = config
        self._disconnect_data = []
        self._width = self._height = self._pixels = None
        if call_mpl_init:
            self._plt = plot.mpl_init(self.config.font_size)

        if fig is None:
            fig = self._plt.figure(
                figsize=self.config.size_inch, FigureClass=SmartplotFigure)
        else:
            assert isinstance(fig, SmartplotFigure)

        fig.set_smartplot(self)

        self._fig = fig
        self._colorbar_width = 0.0
        self._colorbar_height = 0.0
        self._colorbar_axes = []

        self._dims = dims
        self._dim_index = self._dims.index
        self._ndims = len(dims)
        self._labels = {}
        self._aspects = {}

        self.setup_axes()

        self._view_limits = num.zeros((self._ndims, 2))

        self._view_limits[:, :] = num.nan
        self._last_mpl_view_limits = None

        self._x_dims = [dim.lstrip('-') for dim in x_dims]
        self._x_dims_invert = [dim.startswith('-') for dim in x_dims]

        self._y_dims = [dim.lstrip('-') for dim in y_dims]
        self._y_dims_invert = [dim.startswith('-') for dim in y_dims]

        self._z_dims = [dim.lstrip('-') for dim in z_dims]
        self._z_dims_invert = [dim.startswith('-') for dim in z_dims]

        self._mappables = {}
        self._updating_layout = False

        self._need_update_layout = True
        self._update_geometry()

        for axes in self.axes_list:
            fig.add_axes(axes)
            self._connect(axes, 'xlim_changed', self.lim_changed_handler)
            self._connect(axes, 'ylim_changed', self.lim_changed_handler)

        self._cid_resize = fig.canvas.mpl_connect(
            'resize_event', self.resize_handler)

        self._connect(fig, 'dpi_changed', self.dpi_changed_handler)

        self._lim_changed_depth = 0

    def reset_size(self):
        self._fig.set_size_inches(self.config.size_inch)

    def axes(self, ix, iy):
        if not (isinstance(ix, int) and isinstance(iy, int)):
            ix = self._x_dims.index(ix)
            iy = self._y_dims.index(iy)

        return self._axes[iy][ix]

    def set_color_dim(self, mappable, dim):
        assert dim in self._dims
        self._mappables[mappable] = dim

    def set_aspect(self, ydim, xdim, aspect=1.0):
        self._aspects[ydim, xdim] = aspect

    @property
    def dims(self):
        return self._dims

    @property
    def fig(self):
        return self._fig

    @property
    def axes_list(self):
        axes = []
        for row in self._axes:
            axes.extend(row)
        return axes

    @property
    def axes_bottom_list(self):
        return self._axes[0]

    @property
    def axes_left_list(self):
        return [row[0] for row in self._axes]

    def setup_axes(self):
        rect = [0., 0., 1., 1.]
        nx, ny = self._shape
        axes = []
        for iy in range(ny):
            axes.append([])
            for ix in range(nx):
                axes[-1].append(SmartplotAxes(self.fig, rect))

        self._axes = axes

        for _, _, axes_ in self.iaxes():
            axes_.set_autoscale_on(False)

    def _connect(self, obj, sig, handler):
        cid = obj.callbacks.connect(sig, handler)
        self._disconnect_data.append((obj, cid))

    def _disconnect_all(self):
        for obj, cid in self._disconnect_data:
            obj.callbacks.disconnect(cid)

        self._fig.canvas.mpl_disconnect(self._cid_resize)

    def dpi_changed_handler(self, fig):
        if self._updating_layout:
            return

        self._update_geometry()

    def resize_handler(self, event):
        if self._updating_layout:
            return

        self._update_geometry()

    def lim_changed_handler(self, axes):
        if self._updating_layout:
            return

        current = self._get_mpl_view_limits()
        last = self._last_mpl_view_limits
        if last is None:
            return

        for iy, ix, axes in self.iaxes():
            acurrent = current[iy][ix]
            alast = last[iy][ix]
            if acurrent[0] != alast[0]:
                xdim = self._x_dims[ix]
                logger.debug(
                    'X limits have been changed interactively in subplot '
                    '(%i, %i)' % (ix, iy))
                self.set_lim(xdim, *sorted(acurrent[0]))

            if acurrent[1] != alast[1]:
                ydim = self._y_dims[iy]
                logger.debug(
                    'Y limits have been changed interactively in subplot '
                    '(%i, %i)' % (ix, iy))
                self.set_lim(ydim, *sorted(acurrent[1]))

        self.need_update_layout()

    def _update_geometry(self):
        w, h = self._fig.canvas.get_width_height()
        dp = self.get_device_pixel_ratio()
        p = self.get_pixels_factor() * dp

        if (self._width, self._height, self._pixels) != (w, h, p, dp):
            logger.debug(
                'New figure size: %g x %g, '
                'logical-pixel/point: %g, physical-pixel/logical-pixel: %g' % (
                    w, h, p, dp))

            self._width = w                # logical pixel
            self._height = h               # logical pixel
            self._pixels = p               # logical pixel / point
            self._device_pixel_ratio = dp  # physical / logical
            self.need_update_layout()

    @property
    def margins(self):
        return tuple(
            x * self.config.font_size / self._pixels
            for x in self.config.margins_em)

    @property
    def separator(self):
        return self.config.separator_em * self.config.font_size / self._pixels

    def rect_to_figure_coords(self, rect):
        left, bottom, width, height = rect
        return (
            left / self._width,
            bottom / self._height,
            width / self._width,
            height / self._height)

    def point_to_axes_coords(self, axes, point):
        x, y = point
        aleft, abottom, awidth, aheight = axes.get_position().bounds

        x_fig = x / self._width
        y_fig = y / self._height

        x_axes = (x_fig - aleft) / awidth
        y_axes = (y_fig - abottom) / aheight

        return (x_axes, y_axes)

    def get_pixels_factor(self):
        try:
            r = self._fig.canvas.get_renderer()
            return 1.0 / r.points_to_pixels(1.0)
        except AttributeError:
            return 1.0

    def get_device_pixel_ratio(self):
        try:
            return self._fig.canvas.device_pixel_ratio
        except AttributeError:
            return 1.0

    def make_limits(self, lims):
        a = plot.AutoScaler(space=0.05)
        return a.make_scale(lims)[:2]

    def iaxes(self):
        for iy, row in enumerate(self._axes):
            for ix, axes in enumerate(row):
                yield iy, ix, axes

    def get_data_limits(self):
        dim_to_values = defaultdict(list)
        for iy, ix, axes in self.iaxes():
            dim_to_values[self._y_dims[iy]].extend(
                axes.get_yaxis().get_data_interval())
            dim_to_values[self._x_dims[ix]].extend(
                axes.get_xaxis().get_data_interval())

        for mappable, dim in self._mappables.items():
            dim_to_values[dim].extend(mappable.get_clim())

        lims = num.zeros((self._ndims, 2))
        for idim in range(self._ndims):
            dim = self._dims[idim]
            if dim in dim_to_values:
                vs = num.array(
                    dim_to_values[self._dims[idim]], dtype=float)
                vs = vs[num.isfinite(vs)]
                if vs.size > 0:
                    lims[idim, :] = num.min(vs), num.max(vs)
                else:
                    lims[idim, :] = num.nan, num.nan
            else:
                lims[idim, :] = num.nan, num.nan

        lims[num.logical_not(num.isfinite(lims))] = 0.0
        return lims

    def set_lim(self, dim, vmin, vmax):
        assert vmin <= vmax
        self._view_limits[self._dim_index(dim), :] = vmin, vmax

    def _get_mpl_view_limits(self):
        vl = []
        for row in self._axes:
            vl_row = []
            for axes in row:
                vl_row.append((
                    axes.get_xaxis().get_view_interval().tolist(),
                    axes.get_yaxis().get_view_interval().tolist()))

            vl.append(vl_row)

        return vl

    def _remember_mpl_view_limits(self):
        self._last_mpl_view_limits = self._get_mpl_view_limits()

    def window_xmin(self, x):
        return window_min(
            self._shape[0], self._width,
            self.margins[0], self.margins[2] + self._colorbar_width,
            self.separator, x)

    def window_xmax(self, x):
        return window_max(
            self._shape[0], self._width,
            self.margins[0], self.margins[2] + self._colorbar_width,
            self.separator, x)

    def window_ymin(self, y):
        return window_min(
            self._shape[1], self._height,
            self.margins[3] + self._colorbar_height, self.margins[1],
            self.separator, y)

    def window_ymax(self, y):
        return window_max(
            self._shape[1], self._height,
            self.margins[3] + self._colorbar_height, self.margins[1],
            self.separator, y)

    def need_update_layout(self):
        self._need_update_layout = True

    def _update_layout(self):
        assert not self._updating_layout

        if not self._need_update_layout:
            return

        self._updating_layout = True
        try:
            data_limits = self.get_data_limits()

            limits = num.zeros((self._ndims, 2))
            for idim in range(self._ndims):
                limits[idim, :] = self.make_limits(data_limits[idim, :])

            mask = num.isfinite(self._view_limits)
            limits[mask] = self._view_limits[mask]

            # deltas = limits[:, 1] - limits[:, 0]

            # data_w = deltas[0]
            # data_h = deltas[1]

            ml, mt, mr, mb = self.margins
            mr += self._colorbar_width
            mb += self._colorbar_height
            sw = sh = self.separator

            nx, ny = self._shape

            # data_r = data_h / data_w
            em = self.config.font_size / self._pixels
            w = self._width
            h = self._height
            fig_w_avail = w - mr - ml - (nx-1) * sw
            fig_h_avail = h - mt - mb - (ny-1) * sh

            if fig_w_avail <= 0.0 or fig_h_avail <= 0.0:
                raise NotEnoughSpace()

            x_limits = num.zeros((nx, 2))
            for ix, xdim in enumerate(self._x_dims):
                x_limits[ix, :] = limits[self._dim_index(xdim)]

            y_limits = num.zeros((ny, 2))
            for iy, ydim in enumerate(self._y_dims):
                y_limits[iy, :] = limits[self._dim_index(ydim)]

            def get_aspect(dim1, dim2):
                if (dim2, dim1) in self._aspects:
                    return 1.0/self._aspects[dim2, dim1]

                return self._aspects.get((dim1, dim2), None)

            aspects_xx = []
            for ix1, xdim1 in enumerate(self._x_dims):
                for ix2, xdim2 in enumerate(self._x_dims):
                    aspect = get_aspect(xdim2, xdim1)
                    if aspect:
                        aspects_xx.append((ix1, ix2, aspect))

            aspects_yy = []
            for iy1, ydim1 in enumerate(self._y_dims):
                for iy2, ydim2 in enumerate(self._y_dims):
                    aspect = get_aspect(ydim2, ydim1)
                    if aspect:
                        aspects_yy.append((iy1, iy2, aspect))

            aspects_xy = []
            for iy, ix, axes in self.iaxes():
                xdim = self._x_dims[ix]
                ydim = self._y_dims[iy]
                aspect = get_aspect(ydim, xdim)
                if aspect:
                    aspects_xy.append((ix, iy, aspect))

            (x_limits, y_limits), (aws, ahs) = solve_layout_iterative(
                size=(fig_w_avail, fig_h_avail),
                shape=(nx, ny),
                limits=(x_limits, y_limits),
                aspects=(
                    aspects_xx,
                    aspects_yy,
                    aspects_xy))

            for iy, ix, axes in self.iaxes():
                rect = [
                    ml + num.sum(aws[:ix])+(ix*sw),
                    mb + num.sum(ahs[:iy])+(iy*sh),
                    aws[ix], ahs[iy]]

                axes.set_position(
                    self.rect_to_figure_coords(rect), which='both')

                self.set_label_coords(
                    axes, 'x', [
                        wcenter(rect),
                        self.config.label_offset_em[0]*em
                        + self._colorbar_height])

                self.set_label_coords(
                    axes, 'y', [
                        self.config.label_offset_em[1]*em,
                        hcenter(rect)])

                axes.get_xaxis().set_tick_params(
                    bottom=(iy == 0), top=(iy == ny-1),
                    labelbottom=(iy == 0), labeltop=False)

                axes.get_yaxis().set_tick_params(
                    left=(ix == 0), right=(ix == nx-1),
                    labelleft=(ix == 0), labelright=False)

                istride = -1 if self._x_dims_invert[ix] else 1
                axes.set_xlim(*x_limits[ix, ::istride])
                istride = -1 if self._y_dims_invert[iy] else 1
                axes.set_ylim(*y_limits[iy, ::istride])

                axes.tick_params(
                    axis='x',
                    pad=self.config.tick_label_offset_em[0]*em)

                axes.tick_params(
                    axis='y',
                    pad=self.config.tick_label_offset_em[0]*em)

            self._remember_mpl_view_limits()

            for mappable, dim in self._mappables.items():
                mappable.set_clim(*limits[self._dim_index(dim)])

            # scaler = plot.AutoScaler()

            # aspect tick incs same
            #
            # inc = scaler.make_scale(
            #     [0, min(data_expanded_w, data_expanded_h)],
            #     override_mode='off')[2]
            #
            # for axes in self.axes_list:
            #     axes.set_xlim(*limits[0, :])
            #     axes.set_ylim(*limits[1, :])
            #
            #     tl = MultipleLocator(inc)
            #     axes.get_xaxis().set_major_locator(tl)
            #     tl = MultipleLocator(inc)
            #     axes.get_yaxis().set_major_locator(tl)

            for axes, orientation, position in self._colorbar_axes:
                if orientation == 'horizontal':
                    xmin = self.window_xmin(position[0])
                    xmax = self.window_xmax(position[1])
                    ymin = mb - self._colorbar_height
                    ymax = mb - self._colorbar_height \
                        + self.config.colorbar_width_em * em
                else:
                    ymin = self.window_ymin(position[0])
                    ymax = self.window_ymax(position[1])
                    xmin = w - mr + 2 * sw
                    xmax = w - mr + 2 * sw + self.config.colorbar_width_em * em

                rect = [xmin, ymin, xmax-xmin, ymax-ymin]
                axes.set_position(
                    self.rect_to_figure_coords(rect), which='both')

            for ix, axes in enumerate(self.axes_bottom_list):
                dim = self._x_dims[ix]
                s = self._labels.get(dim, dim)
                axes.set_xlabel(s)

            for iy, axes in enumerate(self.axes_left_list):
                dim = self._y_dims[iy]
                s = self._labels.get(dim, dim)
                axes.set_ylabel(s)

        finally:
            self._updating_layout = False

    def set_label_coords(self, axes, which, point):
        axis = axes.get_xaxis() if which == 'x' else axes.get_yaxis()
        axis.set_label_coords(*self.point_to_axes_coords(axes, point))

    def plot(self, points, *args, **kwargs):
        for iy, row in enumerate(self._axes):
            y = points[:, self._dim_index(self._y_dims[iy])]
            for ix, axes in enumerate(row):
                x = points[:, self._dim_index(self._x_dims[ix])]
                axes.plot(x, y, *args, **kwargs)

    def close(self):
        self._disconnect_all()
        self._plt.close(self._fig)

    def show(self):
        self._plt.show()
        self.reset_size()

    def set_label(self, dim, s):
        # just set attribute, handle in update_layout
        self._labels[dim] = s

    def colorbar(
            self, dim,
            orientation='vertical',
            position=None):

        if dim not in self._dims:
            raise PlotError(
                'dimension "%s" is not defined')

        if orientation not in ('vertical', 'horizontal'):
            raise PlotError(
                'orientation must be "vertical" or "horizontal"')

        mappable = None
        for mappable_, dim_ in self._mappables.items():
            if dim_ == dim:
                if mappable is None:
                    mappable = mappable_
                else:
                    mappable_.set_cmap(mappable.get_cmap())

        if mappable is None:
            raise PlotError(
                'no mappable registered for dimension "%s"' % dim)

        if position is None:
            if orientation == 'vertical':
                position = (0, self._shape[1])
            else:
                position = (0, self._shape[0])

        em = self.config.font_size / self._pixels

        if orientation == 'vertical':
            self._colorbar_width = self.config.colorbar_width_em*em + \
                self.separator * 2.0
        else:
            self._colorbar_height = self.config.colorbar_width_em*em + \
                self.separator + self.margins[3]

        axes = SmartplotAxes(self.fig, [0., 0., 1., 1.])
        self.fig.add_axes(axes)

        self._colorbar_axes.append(
            (axes, orientation, position))

        self.need_update_layout()
        # axes.plot([1], [1])
        label = self._labels.get(dim, dim)
        return colorbar.Colorbar(
            axes, mappable, orientation=orientation, label=label)

    def __call__(self, *args):
        return self.axes(*args)


if __name__ == '__main__':
    import sys
    from pyrockoeost import util

    logging.getLogger('matplotlib').setLevel(logging.WARNING)
    util.setup_logging('smartplot', 'debug')

    iplots = [int(x) for x in sys.argv[1:]]

    if 0 in iplots:
        p = Plot(['x'], ['y'])
        n = 100
        x = num.arange(n) * 2.0
        y = num.random.normal(size=n)
        p(0, 0).plot(x, y, 'o')
        p.show()

    if 1 in iplots:
        p = Plot(['x', 'x'], ['y'])
        n = 100
        x = num.arange(n) * 2.0
        y = num.random.normal(size=n)
        p(0, 0).plot(x, y, 'o')
        x = num.arange(n) * 2.0
        y = num.random.normal(size=n)
        p(1, 0).plot(x, y, 'o')
        p.show()

    if 11 in iplots:
        p = Plot(['x'], ['y'])
        p.set_aspect('y', 'x', 2.0)
        n = 100
        xy = num.random.normal(size=(n, 2))
        p(0, 0).plot(xy[:, 0], xy[:, 1], 'o')
        p.show()

    if 12 in iplots:
        p = Plot(['x', 'x2'], ['y'])
        p.set_aspect('x2', 'x', 2.0)
        p.set_aspect('y', 'x', 2.0)
        n = 100
        xy = num.random.normal(size=(n, 2))
        p(0, 0).plot(xy[:, 0], xy[:, 1], 'o')
        p(1, 0).plot(xy[:, 0], xy[:, 1], 'o')
        p.show()

    if 13 in iplots:
        p = Plot(['x'], ['y', 'y2'])
        p.set_aspect('y2', 'y', 2.0)
        p.set_aspect('y', 'x', 2.0)
        n = 100
        xy = num.random.normal(size=(n, 2))
        p(0, 0).plot(xy[:, 0], xy[:, 1], 'o')
        p(0, 1).plot(xy[:, 0], xy[:, 1], 'o')
        p.show()

    if 2 in iplots:
        p = Plot(['easting', 'depth'], ['northing', 'depth'])

        n = 100

        ned = num.random.normal(size=(n, 3))
        p(0, 0).plot(ned[:, 1], ned[:, 0], 'o')
        p(1, 0).plot(ned[:, 2], ned[:, 0], 'o')
        p(0, 1).plot(ned[:, 1], ned[:, 2], 'o')
        p.show()

    if 3 in iplots:
        p = Plot(['easting', 'depth'], ['-depth', 'northing'])
        p.set_aspect('easting', 'northing', 1.0)
        p.set_aspect('easting', 'depth', 0.5)
        p.set_aspect('northing', 'depth', 0.5)

        n = 100

        ned = num.random.normal(size=(n, 3))
        ned[:, 2] *= 0.25
        p(0, 1).plot(ned[:, 1], ned[:, 0], 'o', color='black')
        p(0, 0).plot(ned[:, 1], ned[:, 2], 'o')
        p(1, 1).plot(ned[:, 2], ned[:, 0], 'o')
        p(1, 0).set_visible(False)
        p.set_lim('depth', 0., 0.2)
        p.show()

    if 5 in iplots:
        p = Plot(['time'], ['northing', 'easting', '-depth'], ['depth'])

        n = 100

        t = num.arange(n)
        xyz = num.random.normal(size=(n, 4))
        xyz[:, 0] *= 0.5

        smap = make_smap('summer')

        p(0, 0).scatter(
            t, xyz[:, 0], c=xyz[:, 2], cmap=smap.cmap, norm=smap.norm)
        p(0, 1).scatter(
            t, xyz[:, 1], c=xyz[:, 2], cmap=smap.cmap, norm=smap.norm)
        p(0, 2).scatter(
            t, xyz[:, 2], c=xyz[:, 2], cmap=smap.cmap, norm=smap.norm)

        p.set_lim('depth', -1., 1.)

        p.set_color_dim(smap, 'depth')

        p.set_aspect('northing', 'easting', 1.0)
        p.set_aspect('northing', 'depth', 1.0)

        p.set_label('time', 'Time [s]')
        p.set_label('depth', 'Depth [km]')
        p.set_label('easting', 'Easting [km]')
        p.set_label('northing', 'Northing [km]')

        p.colorbar('depth')

        p.show()

    if 6 in iplots:
        km = 1000.
        p = Plot(
            ['easting'], ['northing']*3, ['displacement'])

        nn, ne = 50, 40
        n = num.linspace(-5*km, 5*km, nn)
        e = num.linspace(-10*km, 10*km, ne)

        displacement = num.zeros((nn, ne, 3))
        g = num.exp(
            -(n[:, num.newaxis]**2 + e[num.newaxis, :]**2) / (5*km)**2)

        displacement[:, :, 0] = g
        displacement[:, :, 1] = g * 0.5
        displacement[:, :, 2] = -g * 0.2

        for icomp in (0, 1, 2):
            c = p(0, icomp).pcolormesh(
                e/km, n/km, displacement[:, :, icomp], shading='gouraud')
            p.set_color_dim(c, 'displacement')

        p.colorbar('displacement')
        p.set_lim('displacement', -1.0, 1.0)
        p.set_label('easting', 'Easting [km]')
        p.set_label('northing', 'Northing [km]')
        p.set_aspect('northing', 'easting')

        p.set_lim('northing', -5.0, 5.0)
        p.set_lim('easting', -3.0, 3.0)
        p.show()



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

from . import base, selection, database, model, io, client, tool, error, \
    environment, dataset, operators, check

from .base import *  # noqa
from .selection import *  # noqa
from .database import *  # noqa
from .model import *  # noqa
from .io import *  # noqa
from .client import *  # noqa
from .tool import *  # noqa
from .error import *  # noqa
from .environment import *  # noqa
from .dataset import *  # noqa
from .operators import *  # noqa
from .check import *  # noqa

__all__ = base.__all__ + selection.__all__ + database.__all__ \
    + model.__all__ + io.__all__ + client.__all__ + tool.__all__ \
    + error.__all__ + environment.__all__ + dataset.__all__ \
    + operators.__all__ + check.__all__



# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------
'''


Analytical crack solutions for surface displacements and fault dislocations.

'''
import numpy as num
import logging

from pyrockoeost.guts import Float, Object
from pyrockoeost.guts_array import Array

guts_prefix = 'modelling'

logger = logging.getLogger(__name__)


class GriffithCrack(Object):
    '''
    Analytical Griffith crack model.
    '''

    width = Float.T(
        help='Width equals to :math:`2 \\cdot a`.',
        default=1.)

    poisson = Float.T(
        help='Poisson ratio :math:`\\nu`.',
        default=.25)

    shearmod = Float.T(
        help='Shear modulus :math:`\\mu` [Pa].',
        default=1.e9)

    stressdrop = Array.T(
        help='Stress drop array:'
             '[:math:`\\sigma_{3,r} - \\sigma_{3,c}`, '
             ':math:`\\sigma_{2,r} - \\sigma_{2,c}`, '
             ':math:`\\sigma_{1,r} - \\sigma_{1,c}`] :math:`=` '
             '[:math:`\\Delta \\sigma_{strike}, '
             '\\Delta \\sigma_{dip}, \\Delta \\sigma_{tensile}`].',
        default=num.array([0., 0., 0.]))

    @property
    def a(self):
        '''
        Half width of the crack in [m].
        '''

        return self.width / 2.

    def disloc_infinite2d(self, x_obs):
        '''
        Calculation of dislocation at crack surface along x2 axis.

        Follows equations by Pollard and Segall (1987) to calculate
        dislocations for an infinite 2D crack extended in x3 direction,
        opening in x1 direction and the crack extending in x2 direction.

        :param x_obs:
            Observation point coordinates along x2-axis.
            If :math:`x_{obs} < -a` or :math:`x_{obs} > a`, output dislocations
            are zero.
        :type x_obs:
            :py:class:`~numpy.ndarray`: ``(N,)``

        :return:
            Dislocations at each observation point in strike, dip and
            tensile direction.
        :rtype:
            :py:class:`~numpy.ndarray`: ``(N, 3)``
        '''

        if type(x_obs) is not num.ndarray:
            x_obs = num.array(x_obs)

        factor = num.array([2. / self.shearmod])
        factor = num.append(
            factor, num.tile(
                2. * (1. - self.poisson) / self.shearmod, (1, 2)))
        factor[1] *= -1.

        crack_el = (x_obs > -self.a) | (x_obs < self.a)

        disl = num.zeros((x_obs.shape[0], 3))
        disl[crack_el, :] = \
            self.stressdrop * num.sqrt(
            self.a**2 - num.tile(x_obs[crack_el, num.newaxis], (1, 3))**2) * \
            factor

        return disl

    def disloc_circular(self, x_obs):
        '''
        Calculation of dislocation at crack surface along x2 axis.

        Follows equations by Pollard and Segall (1987) to calculate
        displacements for a circulat crack extended in x2 and x3 direction and
        opening in x1 direction.

        :param x_obs:
            Observation point coordinates along axis through crack
            centre. If :math:`x_{obs} < -a` or :math:`x_{obs} > a`, output
            dislocations are zero.
        :type x_obs:
            :py:class:`~numpy.ndarray`: ``(N,)``

        :return:
            Dislocations at each observation point in strike, dip and
            tensile direction.
        :rtype:
            :py:class:`~numpy.ndarray`: ``(N, 3)``
        '''

        if type(x_obs) is not num.ndarray:
            x_obs = num.array(x_obs)

        factor = num.array([4. / (self.shearmod * num.pi)])
        factor = num.append(
            factor, num.tile(
                4. * (1. - self.poisson) / (self.shearmod * num.pi), (1, 2)))
        factor[1] *= -1.

        crack_el = (x_obs > -self.a) | (x_obs < self.a)

        disl = num.zeros((x_obs.shape[0], 3))
        disl[crack_el] = \
            self.stressdrop * num.sqrt(
            self.a**2 - num.tile(x_obs[crack_el, num.newaxis], (1, 3))**2) * \
            factor

        return disl

    def _displ_infinite2d_along_x1(self, x1_obs):
        '''
        Calculation of displacement at crack surface along x1-axis.

        Follows equations by Pollard and Segall (1987) to calculate
        displacements for an infinite 2D crack extended in x3 direction,
        opening in x1 direction and the crack tip in x2 direction.

        :param x1_obs:
            Observation point coordinates along x1-axis.
        :type x1_obs:
            :py:class:`~numpy.ndarray`: ``(N,)``

        :return:
            Displacements at each observation point in strike, dip and
            tensile direction.
        :rtype:
            :py:class:`~numpy.ndarray`: ``(M, 3)``
        '''
        displ = num.zeros((x1_obs.shape[0], 3))

        if self.stressdrop[0] != 0.:
            sign = num.sign(x1_obs)
            x1_ratio = x1_obs / self.a

            displ[:, 0] = (
                num.sqrt((x1_ratio)**2 + 1.) - num.abs(
                    x1_ratio)
            ) * self.stressdrop[0] * self.a / self.shearmod * sign

        return displ

    def _displ_infinite2d_along_x2(self, x2_obs):
        '''
        Calculation of displacement at crack surface along x2-axis.

        Follows equations by Pollard and Segall (1987) to calculate
        displacements for an infinite 2D crack extended in x3 direction,
        opening in x1 direction and the crack tip in x2 direction.

        :param x2_obs:
            Observation point coordinates along x2-axis.
        :type x2_obs:
            :py:class:`~numpy.ndarray`: ``(N,)``

        :return:
            Displacements at each observation point in strike, dip and
            tensile direction.
        :rtype:
            :py:class:`~numpy.ndarray`: ``(N, 3)``
        '''

        crack_el = (x2_obs >= -self.a) & (x2_obs <= self.a)

        displ = num.zeros((x2_obs.shape[0], 3))

        if self.stressdrop[1] != 0.:
            factor = (1. - 2. * self.poisson) / (2. * self.shearmod)

            displ[crack_el, 2] = \
                self.stressdrop[1] * factor * x2_obs[crack_el]

            sign = num.sign(x2_obs)

            displ[~crack_el, 2] = \
                self.stressdrop[1] * factor * self.a * sign[~crack_el] * (
                    num.abs(x2_obs[~crack_el] / self.a) -
                    num.sqrt(x2_obs[~crack_el]**2 / self.a**2 - 1.))

        return displ

    def displ_infinite2d(self, x1_obs, x2_obs):
        '''
        Calculation of displacement at crack surface along different axis.

        Follows equations by Pollard and Segall (1987) to calculate
        displacements for an infinite 2D crack extended in x3 direction,
        opening in x1 direction and the crack tip in x2 direction.

        :param x1_obs:
            Observation point coordinates along x1-axis.
            If :math:`x1_obs = 0.`, displacment is calculated along x2-axis.
        :type x1_obs:
            :py:class:`~numpy.ndarray`: ``(M,)``

        :param x2_obs:
            Observation point coordinates along x2-axis.
            If :math:`x2_obs = 0.`, displacment is calculated along x1-axis.
        :type x2_obs:
            :py:class:`~numpy.ndarray`: ``(N,)``

        :return:
            Displacements at each observation point in strike, dip and
            tensile direction.
        :rtype:
            :py:class:`~numpy.ndarray`: ``(M, 3)`` or ``(N, 3)``
        '''

        if type(x1_obs) is not num.ndarray:
            x1_obs = num.array(x1_obs)
        if type(x2_obs) is not num.ndarray:
            x2_obs = num.array(x2_obs)

        if (x1_obs == 0.).all():
            return self._displ_infinite2d_along_x2(x2_obs)
        elif (x2_obs == 0.).all():
            return self._displ_infinite2d_along_x1(x1_obs)


__all__ = [
    'GriffithCrack']



# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------
'''
Access to the USGS Global Crustal Database.

Simple queries and statistical analysis
'''

import numpy as num
import copy
import logging
from os import path

from pyrockoeost.guts import Object, String, Float, Int
from pyrockoeost.guts_array import Array

from pyrockoeost.cake import LayeredModel, Material
from pyrockoeost.plot.cake_plot import my_model_plot, xscaled, yscaled

from .crustdb_abbr import ageKey, provinceKey, referenceKey, pubYear  # noqa

logger = logging.getLogger('pyrockoeost.dataset.crustdb')
THICKNESS_HALFSPACE = 2

db_url = 'https://mirror.pyrocko.org/gsc20130501.txt'
km = 1e3
vel_labels = {
    'vp': '$V_P$',
    'p': '$V_P$',
    'vs': '$V_S$',
    's': '$V_S$',
}


class DatabaseError(Exception):
    pass


class ProfileEmpty(Exception):
    pass


def _getCanvas(axes):

    import matplotlib.pyplot as plt

    if axes is None:
        fig = plt.figure()
        return fig, fig.gca()
    return axes.figure, axes


def xoffset_scale(offset, scale, ax):
    from matplotlib.ticker import ScalarFormatter, AutoLocator

    class FormatVelocities(ScalarFormatter):
        @staticmethod
        def __call__(value, pos):
            return u'%.1f' % ((value-offset) * scale)

    class OffsetLocator(AutoLocator):
        def tick_values(self, vmin, vmax):
            return [v + offset for v in
                    AutoLocator.tick_values(self, vmin, vmax)]

    ax.get_xaxis().set_major_formatter(FormatVelocities())
    ax.get_xaxis().set_major_locator(OffsetLocator())


class VelocityProfile(Object):
    uid = Int.T(
        optional=True,
        help='Unique ID of measurement')

    lat = Float.T(
        help='Latitude [deg]')
    lon = Float.T(
        help='Longitude [deg]')
    elevation = Float.T(
        default=num.nan,
        help='Elevation [m]')
    vp = Array.T(
        shape=(None, 1),
        help='P Wave velocities [m/s]')
    vs = Array.T(
        shape=(None, 1),
        help='S Wave velocities [m/s]')
    d = Array.T(
        shape=(None, 1),
        help='Interface depth, top [m]')
    h = Array.T(
        shape=(None, 1),
        help='Interface thickness [m]')

    heatflow = Float.T(
        optional=True,
        help='Heatflow [W/m^2]')
    geographical_location = String.T(
        optional=True,
        help='Geographic Location')
    geological_province = String.T(
        optional=True,
        help='Geological Province')
    geological_age = String.T(
        optional=True,
        help='Geological Age')
    measurement_method = Int.T(
        optional=True,
        help='Measurement method')
    publication_reference = String.T(
        optional=True,
        help='Publication Reference')
    publication_year__ = Int.T(
        help='Publication Date')

    def __init__(self, *args, **kwargs):
        Object.__init__(self, *args, **kwargs)

        self.h = num.abs(self.d - num.roll(self.d, -1))
        self.h[-1] = 0
        self.nlayers = self.h.size

        self.geographical_location = '%s (%s)' % (
            provinceKey(self.geographical_location),
            self.geographical_location)

        self.vs[self.vs == 0] = num.nan
        self.vp[self.vp == 0] = num.nan

        self._step_vp = num.repeat(self.vp, 2)
        self._step_vs = num.repeat(self.vs, 2)
        self._step_d = num.roll(num.repeat(self.d, 2), -1)
        self._step_d[-1] = self._step_d[-2] + THICKNESS_HALFSPACE

    @property
    def publication_year__(self):
        return pubYear(self.publication_reference)

    def interpolateProfile(self, depths, phase='p', stepped=True):
        '''
        Get a continuous velocity function at arbitrary depth.

        :param depth: Depths to interpolate
        :type depth: :class:`numpy.ndarray`
        :param phase: P or S wave velocity, **p** or **s**
        :type phase: str, optional
        :param stepped: Use a stepped velocity function or gradient
        :type stepped: bool
        :returns: velocities at requested depths
        :rtype: :class:`numpy.ndarray`
        '''

        if phase not in ['s', 'p']:
            raise AttributeError("Phase has to be either 'p' or 's'.")

        if phase == 'p':
            vel = self._step_vp if stepped else self.vp
        elif phase == 's':
            vel = self._step_vs if stepped else self.vs
        d = self._step_d if stepped else self.d

        if vel.size == 0:
            raise ProfileEmpty('Phase %s does not contain velocities' % phase)

        try:
            res = num.interp(depths, d, vel,
                             left=num.nan, right=num.nan)
        except ValueError:
            raise ValueError('Could not interpolate velocity profile.')

        return res

    def plot(self, axes=None):
        '''
        Plot the velocity profile, see :class:`pyrockoeost.cake`.

        :param axes: Axes to plot into.
        :type axes: :class:`matplotlib.Axes`
        '''

        import matplotlib.pyplot as plt

        fig, ax = _getCanvas(axes)
        my_model_plot(self.getLayeredModel(), axes=axes)
        ax.set_title('Global Crustal Database\n'
                     'Velocity Structure at {p.lat:.4f}N, '
                     ' {p.lat:.4f}E (uid {p.uid})'.format(p=self))
        if axes is None:
            plt.show()

    def getLayeredModel(self):
        '''
        Get a layered model, see :class:`pyrockoeost.cake.LayeredModel`.
        '''
        def iterLines():
            for il, m in enumerate(self.iterLayers()):
                yield self.d[il], m, ''

        return LayeredModel.from_scanlines(iterLines())

    def iterLayers(self):
        '''
        Iterator yielding a :class:`pyrockoeost.cake.Material` for each layer.
        '''
        for il in range(self.nlayers):
            yield Material(vp=self.vp[il],
                           vs=self.vs[il])

    @property
    def geog_loc_long(self):
        return provinceKey(self.geog_loc)

    @property
    def geol_age_long(self):
        return ageKey(self.geol_age)

    @property
    def has_s(self):
        return num.any(self.vp)

    @property
    def has_p(self):
        return num.any(self.vs)

    def get_weeded(self):
        '''
        Get weeded representation of layers used in the profile.
        See :func:`pyrockoeost.cake.get_weeded` for details.
        '''
        weeded = num.zeros((self.nlayers, 4))
        weeded[:, 0] = self.d
        weeded[:, 1] = self.vp
        weeded[:, 2] = self.vs

    def _csv(self):
        output = ''
        for d in range(len(self.h)):
            output += (
                    '{p.uid}, {p.lat}, {p.lon},'
                    ' {vp}, {vs}, {h}, {d}, {p.publication_reference}\n'
            ).format(
                p=self,
                vp=self.vp[d], vs=self.vs[d], h=self.h[d], d=self.d[d])
        return output


class CrustDB(object):
    '''
    CrustDB is a container for :class:`VelocityProfile` and provides
    functions for spatial selection, querying, processing and visualising
    data from the Global Crustal Database.
    '''

    def __init__(self, database_file=None, parent=None):
        self.profiles = []
        self._velocity_matrix_cache = {}
        self.data_matrix = None
        self.name = None
        self.database_file = database_file

        if parent is not None:
            pass
        elif database_file is not None:
            self._read(database_file)
        else:
            self._read(self._getRepositoryDatabase())

    def __len__(self):
        return len(self.profiles)

    def __setitem__(self, key, value):
        if not isinstance(value, VelocityProfile):
            raise TypeError('Element is not a VelocityProfile')
        self.profiles[key] = value

    def __delitem__(self, key):
        self.profiles.remove(key)

    def __getitem__(self, key):
        return self.profiles[key]

    def __str__(self):
        rstr = 'Container contains %d velocity profiles:\n\n' % self.nprofiles
        return rstr

    @property
    def nprofiles(self):
        return len(self.profiles)

    def append(self, value):
        if not isinstance(value, VelocityProfile):
            raise TypeError('Element is not a VelocityProfile')
        self.profiles.append(value)

    def copy(self):
        return copy.deepcopy(self)

    def lats(self):
        return num.array(
            [p.lat for p in self.profiles])

    def lons(self):
        return num.array(
            [p.lon for p in self.profiles])

    def _dataMatrix(self):
        if self.data_matrix is not None:
            return self.data_matrix

        self.data_matrix = num.core.records.fromarrays(
            num.vstack([
                num.concatenate([p.vp for p in self.profiles]),
                num.concatenate([p.vs for p in self.profiles]),
                num.concatenate([p.h for p in self.profiles]),
                num.concatenate([p.d for p in self.profiles])
            ]),
            names='vp, vs, h, d')
        return self.data_matrix

    def velocityMatrix(self, depth_range=(0, 60000.), ddepth=100., phase='p'):
        '''
        Create a regular sampled velocity matrix.

        :param depth_range: Depth range, ``(dmin, dmax)``,
            defaults to ``(0, 6000.)``
        :type depth_range: tuple
        :param ddepth: Stepping in [m], defaults to ``100.``
        :type ddepth: float
        :param phase: Phase to calculate ``p`` or ``s``,
            defaults to ``p``
        :type phase: str
        :returns: Sample depths, veloctiy matrix
        :rtype: tuple, (sample_depth, :class:`numpy.ndarray`)
        '''
        dmin, dmax = depth_range
        uid = '.'.join(map(repr, (dmin, dmax, ddepth, phase)))
        sdepth = num.linspace(dmin, dmax, (dmax - dmin) / ddepth)
        ndepth = sdepth.size

        if uid not in self._velocity_matrix_cache:
            vel_mat = num.empty((self.nprofiles, ndepth))
            for ip, profile in enumerate(self.profiles):
                vel_mat[ip, :] = profile.interpolateProfile(sdepth,
                                                            phase=phase)
            self._velocity_matrix_cache[uid] = num.ma.masked_invalid(vel_mat)

        return sdepth, self._velocity_matrix_cache[uid]

    def rmsRank(self, ref_profile, depth_range=(0, 3500.), ddepth=100.,
                phase='p'):
        '''
        Correlates ``ref_profile`` to each profile in the database.

        :param ref_profile: Reference profile
        :type ref_profile: :class:`VelocityProfile`
        :param depth_range: Depth range in [m], ``(dmin, dmax)``,
            defaults to ``(0, 35000.)``
        :type depth_range: tuple, optional
        :param ddepth: Stepping in [m], defaults to ``100.``
        :type ddepth: float
        :param phase: Phase to calculate ``p`` or ``s``, defaults to ``p``
        :type phase: str
        :returns: RMS factor length of N_profiles
        :rtype: :class:`numpy.ndarray`
        '''
        if not isinstance(ref_profile, VelocityProfile):
            raise ValueError('ref_profile is not a VelocityProfile')

        sdepth, vel_matrix = self.velocityMatrix(depth_range, ddepth,
                                                 phase=phase)
        ref_vel = ref_profile.interpolateProfile(sdepth, phase=phase)

        rms = num.empty(self.nprofiles)
        for p in range(self.nprofiles):
            profile = vel_matrix[p, :]
            rms[p] = num.sqrt(profile**2 - ref_vel**2).sum() / ref_vel.size
        return rms

    def histogram2d(self, depth_range=(0., 60000.), vel_range=None,
                    ddepth=100., dvbin=100., ddbin=2000., phase='p'):
        '''
        Create a 2D Histogram of all the velocity profiles.

        Check :func:`numpy.histogram2d` for more information.

        :param depth_range: Depth range in [m], ``(dmin, dmax)``,
            defaults to ``(0., 60000.)``
        :type depth_range: tuple
        :param vel_range: Depth range, ``(vmin, vmax)``,
            defaults to ``(5500., 8500.)``
        :type vel_range: tuple
        :param ddepth: Stepping in [km], defaults to ``100.``
        :type ddepth: float
        :param dvbin: Bin size in velocity dimension [m/s], defaults to 100.
        :type dvbin: float
        :param dvbin: Bin size in depth dimension [m], defaults to 2000.
        :type dvbin: float
        :param phase: Phase to calculate ``p`` or ``s``, defaults to ``p``
        :type phase: str

        :returns: :func:`numpy.histogram2d`
        :rtype: tuple
        '''
        sdepth, v_mat = self.velocityMatrix(depth_range, ddepth, phase=phase)
        d_vec = num.tile(sdepth, self.nprofiles)

        # Velocity and depth bins
        if vel_range is None:
            vel_range = ((v_mat.min() // 1e2) * 1e2,
                         (v_mat.max() // 1e2) * 1e2)
        nvbins = int((vel_range[1] - vel_range[0]) / dvbin)
        ndbins = int((depth_range[1] - depth_range[0]) / ddbin)

        return num.histogram2d(v_mat.flatten(), d_vec,
                               range=(vel_range, depth_range),
                               bins=[nvbins, ndbins],
                               normed=False)

    def meanVelocity(self, depth_range=(0., 60000.), ddepth=100., phase='p'):
        '''
        Get mean and standard deviation of velocity profile.

        :param depth_range: Depth range in [m], ``(dmin, dmax)``,
            defaults to ``(0., 60000.)``
        :type depth_range: tuple
        :param ddepth: Stepping in [m], defaults to ``100.``
        :type ddepth: float
        :param phase: Phase to calculate ``p`` or ``s``, defaults to ``p``
        :type phase: str
        :returns: depth vector, mean velocities, standard deviations
        :rtype: tuple of :class:`numpy.ndarray`
        '''
        sdepth, v_mat = self.velocityMatrix(depth_range, ddepth, phase=phase)
        v_mean = num.ma.mean(v_mat, axis=0)
        v_std = num.ma.std(v_mat, axis=0)

        return sdepth, v_mean.flatten(), v_std.flatten()

    def modeVelocity(self, depth_range=(0., 60000.), ddepth=100., phase='p'):
        '''
        Get mode velocity profile and standard deviation.

        :param depth_range: Depth range in [m], ``(dmin, dmax)``,
            defaults to ``(0., 60000.)``
        :type depth_range: tuple
        :param ddepth: Stepping in [m], defaults to ``100.``
        :type ddepth: float
        :param phase: Phase to calculate ``p`` or ``s``, defaults to ``p``
        :type phase: str
        :returns: depth vector, mode velocity, number of counts at each depth
        :rtype: tuple of :class:`numpy.ndarray`
        '''
        import scipy.stats

        sdepth, v_mat = self.velocityMatrix(depth_range, ddepth)
        v_mode, v_counts = scipy.stats.mstats.mode(v_mat, axis=0)
        return sdepth, v_mode.flatten(), v_counts.flatten()

    def medianVelocity(self, depth_range=(0., 60000.), ddepth=100., phase='p'):
        '''
        Median velocity profile plus std variation.

        :param depth_range: Depth range in [m], ``(dmin, dmax)``,
            defaults to ``(0., 60000.)``
        :type depth_range: tuple
        :param ddepth: Stepping in [m], defaults to ``100.``
        :type ddepth: float
        :param phase: Phase to calculate ``p`` or ``s``, defaults to ``p``
        :type phase: str
        :returns: depth vector, median velocities, standard deviations
        :rtype: tuple of :class:`numpy.ndarray`
        '''
        sdepth, v_mat = self.velocityMatrix(depth_range, ddepth, phase=phase)
        v_mean = num.ma.median(v_mat, axis=0)
        v_std = num.ma.std(v_mat, axis=0)

        return sdepth, v_mean.flatten(), v_std.flatten()

    def plotHistogram(self, vel_range=None, bins=36, phase='vp',
                      axes=None):
        '''
        Plot 1D histogram of seismic velocities in the container.

        :param vel_range: Velocity range, defaults to (5.5, 8.5)
        :type vel_range: tuple, optional
        :param bins: bins, defaults to 30 (see :func:`numpy.histogram`)
        :type bins: int, optional
        :param phase: Property to plot out of ``['vp', 'vs']``,
            defaults to 'vp'
        :type phase: str, optional
        :param figure: Figure to plot in, defaults to None
        :type figure: :class:`matplotlib.Figure`, optional
        '''

        import matplotlib.pyplot as plt

        fig, ax = _getCanvas(axes)

        if phase not in ['vp', 'vs']:
            raise AttributeError('phase has to be either vp or vs')

        data = self._dataMatrix()[phase]

        ax.hist(data, weights=self.data_matrix['h'],
                range=vel_range, bins=bins,
                color='g', alpha=.5)
        ax.text(.95, .95, '%d Profiles' % self.nprofiles,
                transform=ax.transAxes, fontsize=10,
                va='top', ha='right', alpha=.7)

        ax.set_title('Distribution of %s' % vel_labels[phase])
        ax.set_xlabel('%s [km/s]' % vel_labels[phase])
        ax.set_ylabel('Cumulative occurrence [N]')
        xscaled(1./km, ax)
        ax.yaxis.grid(alpha=.4)

        if self.name is not None:
            ax.set_title('%s for %s' % (ax.get_title(), self.name))

        if axes is None:
            plt.show()

    def plot(self, depth_range=(0, 60000.), ddepth=100., ddbin=2000.,
             vel_range=None, dvbin=100.,
             percent=False,
             plot_mode=True, plot_median=True, plot_mean=False,
             show_cbar=True,
             aspect=.02,
             phase='p',
             axes=None):
        '''
        Plot a two 2D Histogram of seismic velocities.

        :param depth_range: Depth range, ``(dmin, dmax)``,
            defaults to ``(0, 60)``
        :type depth_range: tuple
        :param vel_range: Velocity range, ``(vmin, vmax)``
        :type vel_range: tuple
        :param ddepth: Stepping in [m], defaults to ``.1``
        :type ddepth: float
        :param dvbin: Bin size in velocity dimension [m/s], defaults to .1
        :type dvbin: float
        :param dvbin: Bin size in depth dimension [m], defaults to 2000.
        :type dvbin: float
        :param phase: Phase to calculate ``p`` or ``s``, defaults to ``p``
        :type phase: str
        :param plot_mode: Plot the Mode
        :type plot_mode: bool
        :param plot_mean: Plot the Mean
        :type plot_mean: bool
        :param plot_median: Plot the Median
        :type plot_median: bool
        :param axes: Axes to plot into, defaults to None
        :type axes: :class:`matplotlib.Axes`
        '''

        import matplotlib.pyplot as plt

        fig, ax = _getCanvas(axes)

        ax = fig.gca()

        if vel_range is not None:
            vmin, vmax = vel_range
        dmin, dmax = depth_range

        vfield, vedg, dedg = self.histogram2d(vel_range=vel_range,
                                              depth_range=depth_range,
                                              ddepth=ddepth, dvbin=dvbin,
                                              ddbin=ddbin, phase=phase)
        vfield /= (ddbin / ddepth)

        if percent:
            vfield /= vfield.sum(axis=1)[num.newaxis, :]

        grid_ext = [vedg[0], vedg[-1], dedg[-1], dedg[0]]
        histogram = ax.imshow(vfield.swapaxes(0, 1),
                              interpolation='nearest',
                              extent=grid_ext, aspect=aspect)

        if show_cbar:
            cticks = num.unique(
                num.arange(0, vfield.max(), vfield.max() // 10).round())
            cbar = fig.colorbar(histogram, ticks=cticks, format='%1i',
                                orientation='horizontal')
            if percent:
                cbar.set_label('Percent')
            else:
                cbar.set_label('Number of Profiles')

        if plot_mode:
            sdepth, vel_mode, _ = self.modeVelocity(depth_range=depth_range,
                                                    ddepth=ddepth)
            ax.plot(vel_mode[sdepth < dmax] + ddepth/2,
                    sdepth[sdepth < dmax],
                    alpha=.8, color='w', label='Mode')

        if plot_mean:
            sdepth, vel_mean, _ = self.meanVelocity(depth_range=depth_range,
                                                    ddepth=ddepth)
            ax.plot(vel_mean[sdepth < dmax] + ddepth/2,
                    sdepth[sdepth < dmax],
                    alpha=.8, color='w', linestyle='--', label='Mean')

        if plot_median:
            sdepth, vel_median, _ = self.medianVelocity(
                                        depth_range=depth_range,
                                        ddepth=ddepth)
            ax.plot(vel_median[sdepth < dmax] + ddepth/2,
                    sdepth[sdepth < dmax],
                    alpha=.8, color='w', linestyle=':', label='Median')

        ax.grid(True, which='both', color='w', linewidth=.8, alpha=.4)

        ax.text(.025, .025, '%d Profiles' % self.nprofiles,
                color='w', alpha=.7,
                transform=ax.transAxes, fontsize=9, va='bottom', ha='left')

        ax.set_title('Crustal Velocity Distribution')
        ax.set_xlabel('%s [km/s]' % vel_labels[phase])
        ax.set_ylabel('Depth [km]')
        yscaled(1./km, ax)
        xoffset_scale(dvbin/2, 1./km, ax)
        ax.set_xlim(vel_range)

        if self.name is not None:
            ax.set_title('%s for %s' % (ax.get_title(), self.name))

        if plot_mode or plot_mean or plot_median:
            leg = ax.legend(loc=1, fancybox=True, prop={'size': 10.})
            leg.get_frame().set_alpha(.6)

        if axes is None:
            plt.show()

    def plotVelocitySurface(self, v_max, d_min=0., d_max=6000., axes=None):
        '''
        Plot a triangulated a depth surface exceeding velocity.
        '''

        import matplotlib.pyplot as plt

        fig, ax = _getCanvas(axes)
        d = self.exceedVelocity(v_max, d_min, d_max)
        lons = self.lons()[d > 0]
        lats = self.lats()[d > 0]
        d = d[d > 0]

        ax.tricontourf(lats, lons, d)

        if axes is None:
            plt.show()

    def plotMap(self, outfile, **kwargs):
        from pyrockoeost.plot import gmtpy
        lats = self.lats()
        lons = self.lons()
        s, n, w, e = (lats.min(), lats.max(), lons.min(), lons.max())

        def darken(c, f=0.7):
            return (c[0]*f, c[1]*f, c[2]*f)

        gmt = gmtpy.GMT()
        gmt.psbasemap(B='40/20',
                      J='M0/12',
                      R='%f/%f/%f/%f' % (w, e, s, n))
        gmt.pscoast(R=True, J=True,
                    D='i', S='216/242/254', A=10000,
                    W='.2p')
        gmt.psxy(R=True, J=True,
                 in_columns=[lons, lats],
                 S='c2p', G='black')
        gmt.save(outfile)

    def exceedVelocity(self, v_max, d_min=0, d_max=60):
        '''
        Returns the last depth ``v_max`` has not been exceeded.

        :param v_max: maximal velocity
        :type vmax: float
        :param dz: depth is sampled in dz steps
        :type dz: float
        :param d_max: maximum depth
        :type d_max: int
        :param d_min: minimum depth
        :type d_min: int
        :returns: Lat, Lon, Depth and uid where ``v_max`` is exceeded
        :rtype: list(num.array)
        '''
        self.profile_exceed_velocity = num.empty(len(self.profiles))
        self.profile_exceed_velocity[:] = num.nan

        for ip, profile in enumerate(self.profiles):
            for il in range(len(profile.d)):
                if profile.d[il] <= d_min\
                        or profile.d[il] >= d_max:
                    continue
                if profile.vp[il] < v_max:
                    continue
                else:
                    self.profile_exceed_velocity[ip] = profile.d[il]
                    break
        return self.profile_exceed_velocity

    def selectRegion(self, west, east, south, north):
        '''
        Select profiles within a region by geographic corner coordinates.

        :param west: west corner
        :type west: float
        :param east: east corner
        :type east: float
        :param south: south corner
        :type south: float
        :param north: north corner
        :type north: float
        :returns: Selected profiles
        :rtype: :class:`CrustDB`
        '''
        r_container = self._emptyCopy()

        for profile in self.profiles:
            if profile.lon >= west and profile.lon <= east \
                    and profile.lat <= north and profile.lat >= south:
                r_container.append(profile)

        return r_container

    def selectPolygon(self, poly):
        '''
        Select profiles within a polygon.

        The algorithm is called the **Ray Casting Method**

        :param poly: Latitude Longitude pairs of the polygon
        :type param: list of :class:`numpy.ndarray`
        :returns: Selected profiles
        :rtype: :class:`CrustDB`
        '''
        r_container = self._emptyCopy()

        for profile in self.profiles:
            x = profile.lon
            y = profile.lat

            inside = False
            p1x, p1y = poly[0]
            for p2x, p2y in poly:
                if y >= min(p1y, p2y):
                    if y <= max(p1y, p2y):
                        if x <= max(p1x, p2x):
                            if p1y != p2y:
                                xints = (y - p1y) * (p2x - p1x) / \
                                    (p2y - p1y) + p1x
                            if p1x == p2x or x <= xints:
                                inside = not inside
                p1x, p1y = p2x, p2y
            if inside:
                r_container.append(profile)

        return r_container

    def selectLocation(self, lat, lon, radius=10):
        '''
        Select profiles at a geographic location within a ``radius``.

        :param lat: Latitude in [deg]
        :type lat: float
        :param lon: Longitude in [deg]
        :type lon: float
        :param radius: Radius in [deg]
        :type radius: float
        :returns: Selected profiles
        :rtype: :class:`CrustDB`
        '''
        r_container = self._emptyCopy()
        logger.info('Selecting location %f, %f (r=%f)...' % (lat, lon, radius))
        for profile in self.profiles:
            if num.sqrt((lat - profile.lat)**2 +
                        (lon - profile.lon)**2) <= radius:
                r_container.append(profile)

        return r_container

    def selectMinLayers(self, nlayers):
        '''
        Select profiles with more than ``nlayers``.

        :param nlayers: Minimum number of layers
        :type nlayers: int
        :returns: Selected profiles
        :rtype: :class:`CrustDB`
        '''
        r_container = self._emptyCopy()
        logger.info('Selecting minimum %d layers...' % nlayers)

        for profile in self.profiles:
            if profile.nlayers >= nlayers:
                r_container.append(profile)

        return r_container

    def selectMaxLayers(self, nlayers):
        '''
        Select profiles with more than ``nlayers``.

        :param nlayers: Maximum number of layers
        :type nlayers: int
        :returns: Selected profiles
        :rtype: :class:`CrustDB`
        '''
        r_container = self._emptyCopy()
        logger.info('Selecting maximum %d layers...' % nlayers)

        for profile in self.profiles:
            if profile.nlayers <= nlayers:
                r_container.append(profile)

        return r_container

    def selectMinDepth(self, depth):
        '''
        Select profiles describing layers deeper than ``depth``.

        :param depth: Minumum depth in [m]
        :type depth: float
        :returns: Selected profiles
        :rtype: :class:`CrustDB`
        '''
        r_container = self._emptyCopy()
        logger.info('Selecting minimum depth %f m...' % depth)

        for profile in self.profiles:
            if profile.d.max() >= depth:
                r_container.append(profile)
        return r_container

    def selectMaxDepth(self, depth):
        '''
        Select profiles describing layers shallower than ``depth``.

        :param depth: Maximum depth in [m]
        :type depth: float
        :returns: Selected profiles
        :rtype: :class:`CrustDB`
        '''
        r_container = self._emptyCopy()
        logger.info('Selecting maximum depth %f m...' % depth)

        for profile in self.profiles:
            if profile.d.max() <= depth:
                r_container.append(profile)
        return r_container

    def selectVp(self):
        '''
        Select profiles describing P wave velocity.

        :returns Selected profiles
        :rtype: :class:`CrustDB`
        '''
        r_container = self._emptyCopy()
        logger.info('Selecting profiles providing Vp...')

        for profile in self.profiles:
            if not num.all(num.isnan(profile.vp)):
                r_container.append(profile)
        return r_container

    def selectVs(self):
        '''
        Select profiles describing P wave velocity.

        :returns: Selected profiles
        :rtype: :class:`CrustDB`
        '''
        r_container = self._emptyCopy()
        logger.info('Selecting profiles providing Vs...')

        for profile in self.profiles:
            if not num.all(num.isnan(profile.vs)):
                r_container.append(profile)
        return r_container

    def _emptyCopy(self):
        r_container = CrustDB(parent=self)
        r_container.name = self.name
        return r_container

    def exportCSV(self, filename=None):
        '''
        Export as CSV file.

        :param filename: Export filename
        :type filename: str
        '''
        with open(filename, 'w') as file:
            file.write('# uid, Lat, Lon, vp, vs, H, Depth, Reference\n')
            for profile in self.profiles:
                file.write(profile._csv())

    def exportYAML(self, filename=None):
        '''
        Export as YAML file.

        :param filename: Export filename
        :type filename: str
        '''
        with open(filename, 'w') as file:
            for profile in self.profiles:
                file.write(profile.__str__())

    @classmethod
    def readDatabase(cls, database_file):
        db = cls()
        CrustDB._read(db, database_file)
        return db

    @staticmethod
    def _getRepositoryDatabase():
        from pyrockoeost import config

        name = path.basename(db_url)
        data_path = path.join(config.config().crustdb_dir, name)
        if not path.exists(data_path):
            from pyrockoeost import util
            util.download_file(db_url, data_path, None, None)

        return data_path

    def _read(self, database_file):
        '''
        Reads in the the GSN database file and puts it in CrustDB.

        File format:

   uid  lat/lon  vp    vs    hc     depth
    2   29.76N   2.30   .00   2.00    .00  s  25.70   .10    .00  NAC-CO   5 U
        96.31W   3.94   .00   5.30   2.00  s  33.00   MCz  39.00  61C.3    EXC
                 5.38   .00  12.50   7.30  c
                 6.92   .00  13.20  19.80  c
                 8.18   .00    .00  33.00  m

    3   34.35N   3.00   .00   3.00    .00  s  35.00  1.60    .00  NAC-BR   4 R
       117.83W   6.30   .00  16.50   3.00     38.00   MCz  55.00  63R.1    ORO
                 7.00   .00  18.50  19.50
                 7.80   .00    .00  38.00  m


        :param database_file: path to database file, type string

        '''

        def get_empty_record():
            meta = {
                'uid': num.nan,
                'geographical_location': None,
                'geological_province': None,
                'geological_age': None,
                'elevation': num.nan,
                'heatflow': num.nan,
                'measurement_method': None,
                'publication_reference': None
            }
            # vp, vs, h, d, lat, lon, meta
            return (num.zeros(128, dtype=num.float32),
                    num.zeros(128, dtype=num.float32),
                    num.zeros(128, dtype=num.float32),
                    num.zeros(128, dtype=num.float32),
                    0., 0., meta)

        nlayers = []

        def add_record(vp, vs, h, d, lat, lon, meta, nlayer):
            if nlayer == 0:
                return
            self.append(VelocityProfile(
                vp=vp[:nlayer] * km,
                vs=vs[:nlayer] * km,
                h=h[:nlayer] * km,
                d=d[:nlayer] * km,
                lat=lat, lon=lon,
                **meta))
            nlayers.append(nlayer)

        vp, vs, h, d, lat, lon, meta = get_empty_record()
        ilayer = 0
        with open(database_file, 'r') as database:
            for line, dbline in enumerate(database):
                if dbline.isspace():
                    if not len(d) == 0:
                        add_record(vp, vs, h, d, lat, lon, meta, ilayer)
                    if not len(vp) == len(h):
                        raise DatabaseError(
                            'Inconsistent database, check line %d!\n\tDebug: '
                            % line, lat, lon, vp, vs, h, d, meta)

                    vp, vs, h, d, lat, lon, meta = get_empty_record()
                    ilayer = 0
                else:
                    try:
                        if ilayer == 0:
                            lat = float(dbline[8:13])
                            if dbline[13] == b'S':
                                lat = -lat
                            # Additional meta data
                            meta['uid'] = int(dbline[0:6])
                            meta['elevation'] = float(dbline[52:57])
                            meta['heatflow'] = float(dbline[58:64])
                            if meta['heatflow'] == 0.:
                                meta['heatflow'] = None
                            meta['geographical_location'] =\
                                dbline[66:72].strip()
                            meta['measurement_method'] = dbline[77]
                        if ilayer == 1:
                            lon = float(dbline[7:13])
                            if dbline[13] == b'W':
                                lon = -lon
                            # Additional meta data
                            meta['geological_age'] = dbline[54:58].strip()
                            meta['publication_reference'] =\
                                dbline[66:72].strip()
                            meta['geological_province'] = dbline[74:78].strip()
                        try:
                            vp[ilayer] = dbline[17:21]
                            vs[ilayer] = dbline[23:27]
                            h[ilayer] = dbline[28:34]
                            d[ilayer] = dbline[35:41]
                        except ValueError:
                            pass
                    except ValueError:
                        logger.warning(
                            'Could not interpret line %d, skipping\n%s' %
                            (line, dbline))
                        while not database.readlines():
                            pass
                        vp, vs, h, d, lat, lon, meta = get_empty_record()
                    ilayer += 1
            # Append last profile
            add_record(vp, vs, h, d, lat, lon, meta, ilayer)
            logger.info('Loaded %d profiles from %s' %
                        (self.nprofiles, database_file))



# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import math

from pyrockoeost.guts import Int, StringChoice, Bool, Float
from pyrockoeost.gui.qt_compat import qw, qc

from .. import common
from pyrockoeost.gui.vtk_util import TrimeshPipe
from .base import Element, ElementState
from pyrockoeost import icosphere
from pyrockoeost.geometry import r2d
from pyrockoeost.color import Color

km = 1000.

guts_prefix = 'sparrow'


class IcosphereBaseChoice(StringChoice):
    choices = ['icosahedron', 'tetrahedron', 'tcube']


class IcosphereKindChoice(StringChoice):
    choices = ['kind1', 'kind2']


class IcosphereState(ElementState):
    base = IcosphereBaseChoice.T(default='icosahedron')
    kind = IcosphereKindChoice.T(default='kind1')
    level = Int.T(default=0)
    visible = Bool.T(default=True)
    smooth = Bool.T(default=False)
    color = Color.T(default=Color.D('aluminium5'))

    ambient = Float.T(default=0.0)
    diffuse = Float.T(default=1.0)
    specular = Float.T(default=0.0)

    opacity = Float.T(default=1.0)
    depth = Float.T(default=30.0*km)

    def create(self):
        element = IcosphereElement()
        return element


class IcosphereElement(Element):

    def __init__(self):
        Element.__init__(self)
        self._mesh = None
        self._controls = None
        self._params = None

    def get_name(self):
        return 'Icosphere'

    def bind_state(self, state):
        Element.bind_state(self, state)
        variables = [
            'visible', 'level', 'base', 'kind', 'smooth', 'color', 'ambient',
            'diffuse', 'specular', 'depth', 'opacity']

        self.talkie_connect(state, variables, self.update)

    def set_parent(self, parent):
        Element.set_parent(self, parent)
        self.talkie_connect(self._parent.state, 'dip', self.update)

        self._parent.add_panel(
            self.get_title_label(),
            self._get_controls(),
            visible=True,
            title_controls=[
                self.get_title_control_remove(),
                self.get_title_control_visible()])

        self.update()

    def unset_parent(self):
        self.unbind_state()
        if self._parent:
            if self._mesh:
                self._parent.remove_actor(self._mesh.actor)
                self._mesh = None

            if self._controls:
                self._parent.remove_panel(self._controls)
                self._controls = None

            self._parent.update_view()
            self._parent = None

    def update(self, *args):
        state = self._state

        params = state.level, state.base, state.kind, state.smooth

        if self._mesh and (params != self._params or not state.visible):
            self._parent.remove_actor(self._mesh.actor)
            self._mesh = None

        if state.visible and not self._mesh:
            vertices, faces = icosphere.sphere(
                state.level, state.base, state.kind, radius=1.0,
                triangulate=False)

            self._vertices = vertices
            self._depth = 0.0

            self._mesh = TrimeshPipe(vertices, faces, smooth=state.smooth)
            self._params = params

            self._parent.add_actor(self._mesh.actor)

        if self._parent.state.distance < 2.0:
            angle = 180. - math.acos(self._parent.state.distance / 2.0)*r2d

            opacity = min(
                1.0,
                max(0., (angle+5. - self._parent.state.dip) / 10.))
        else:
            opacity = 1.0

        opacity *= state.opacity

        if self._mesh:
            if self._depth != state.depth:
                radius = (self._parent.planet_radius - state.depth) \
                    / self._parent.planet_radius

                self._mesh.set_vertices(self._vertices * radius)
                self._depth = state.depth

            self._mesh.set_opacity(opacity)
            self._mesh.set_color(state.color)
            self._mesh.set_ambient(state.ambient)
            self._mesh.set_diffuse(state.diffuse)
            self._mesh.set_specular(state.specular)

        self._parent.update_view()

    def _get_controls(self):
        state = self._state
        if not self._controls:
            from ..state import state_bind_slider, \
                state_bind_combobox, state_bind_checkbox, \
                state_bind_combobox_color, state_bind_lineedit

            frame = qw.QFrame()
            layout = qw.QGridLayout()
            frame.setLayout(layout)

            slider = qw.QSlider(qc.Qt.Horizontal)
            slider.setSizePolicy(
                qw.QSizePolicy(
                    qw.QSizePolicy.Expanding, qw.QSizePolicy.Fixed))
            slider.setMinimum(0)
            slider.setMaximum(5)
            slider.setSingleStep(1)
            slider.setPageStep(1)

            layout.addWidget(qw.QLabel('Level'), 0, 0)
            layout.addWidget(slider, 0, 1)

            state_bind_slider(self, state, 'level', slider, dtype=int)

            cb = common.string_choices_to_combobox(IcosphereBaseChoice)
            layout.addWidget(qw.QLabel('Base'), 1, 0)
            layout.addWidget(cb, 1, 1)
            state_bind_combobox(self, state, 'base', cb)

            cb = common.string_choices_to_combobox(IcosphereKindChoice)
            layout.addWidget(qw.QLabel('Kind'), 2, 0)
            layout.addWidget(cb, 2, 1)
            state_bind_combobox(self, state, 'kind', cb)

            layout.addWidget(qw.QLabel('Color'), 3, 0)

            cb = common.strings_to_combobox(
                ['black', 'aluminium6', 'aluminium5', 'aluminium4',
                 'aluminium3', 'aluminium2', 'aluminium1', 'white',
                 'scarletred2', 'orange2', 'skyblue2', 'plum2'])

            layout.addWidget(cb, 3, 1)
            state_bind_combobox_color(
                self, self._state, 'color', cb)

            def add_slider(title, param, irow):
                layout.addWidget(qw.QLabel(title), irow, 0)

                slider = qw.QSlider(qc.Qt.Horizontal)
                slider.setSizePolicy(
                    qw.QSizePolicy(
                        qw.QSizePolicy.Expanding, qw.QSizePolicy.Fixed))
                slider.setMinimum(0)
                slider.setMaximum(1000)
                layout.addWidget(slider, irow, 1)

                state_bind_slider(
                    self, state, param, slider, factor=0.001)

            add_slider('Opacity', 'opacity', 4)
            add_slider('Ambient', 'ambient', 5)
            add_slider('Diffuse', 'diffuse', 6)
            add_slider('Specular', 'specular', 7)

            cb = qw.QCheckBox('Smooth')
            layout.addWidget(cb, 8, 1)
            state_bind_checkbox(self, state, 'smooth', cb)

            layout.addWidget(qw.QLabel('Depth [km]'), 9, 0)
            le = qw.QLineEdit()
            layout.addWidget(le, 9, 1)
            state_bind_lineedit(
                self, state, 'depth', le,
                from_string=lambda s: float(s)*1000.,
                to_string=lambda v: str(v/1000.))

            layout.addWidget(qw.QFrame(), 10, 0, 1, 2)

        self._controls = frame

        return self._controls


__all__ = [
    'IcosphereElement',
    'IcosphereState',
    'IcosphereKindChoice',
    'IcosphereBaseChoice']
